@@ -1,51 +1,48 @@
-from heapq import (heappush,
-                   heapreplace)
-from operator import attrgetter
-from typing import (Iterator,
-                    List,
-                    Optional,
-                    Sequence,
-                    Tuple,
-                    Type,
-                    Union)
+from heapq import (heappush as _heappush,
+                   heapreplace as _heapreplace)
+from operator import attrgetter as _attrgetter
+from typing import (Callable as _Callable,
+                    Iterator as _Iterator,
+                    List as _List,
+                    Optional as _Optional,
+                    Sequence as _Sequence,
+                    Tuple as _Tuple,
+                    Type as _Type,
+                    Union as _Union)
 
-from ground.hints import (Box,
-                          Coordinate,
-                          Point)
-from reprit.base import generate_repr
+from ground.base import (Context as _Context,
+                         get_context as _get_context)
+from ground.hints import (Box as _Box,
+                          Coordinate as _Coordinate,
+                          Point as _Point)
+from reprit.base import generate_repr as _generate_repr
 
 from .core import box as _box
-from .core.utils import points_distance as _points_distance
 
-try:
-    from typing import Protocol
-except ImportError:
-    from typing_extensions import Protocol
-
-Item = Tuple[int, Point]
+Item = _Tuple[int, _Point]
 NIL = None
 
-_PROJECTORS = attrgetter('x'), attrgetter('y')
+_PROJECTORS = _attrgetter('x'), _attrgetter('y')
 
 
 class Node:
-    """
-    Represents node of *kd*-tree.
-    """
-
-    __slots__ = 'index', 'point', 'is_y_axis', 'projector', 'left', 'right'
+    """Represents node of *kd*-tree."""
+    __slots__ = ('index', 'is_y_axis', 'left', 'metric', 'point', 'projector',
+                 'right')
 
     def __init__(self,
                  index: int,
-                 point: Point,
+                 point: _Point,
                  is_y_axis: bool,
-                 left: Union['Node', NIL],
-                 right: Union['Node', NIL]) -> None:
+                 left: _Union['Node', NIL],
+                 right: _Union['Node', NIL],
+                 metric: _Callable[[_Point, _Point], _Coordinate]) -> None:
         self.index, self.point = index, point
         self.is_y_axis, self.projector = is_y_axis, _PROJECTORS[is_y_axis]
         self.left, self.right = left, right
+        self.metric = metric
 
-    __repr__ = generate_repr(__init__)
+    __repr__ = _generate_repr(__init__)
 
     @property
     def item(self) -> Item:
@@ -53,15 +50,15 @@ class Node:
         return self.index, self.point
 
     @property
-    def projection(self) -> Coordinate:
+    def projection(self) -> _Coordinate:
         """Returns projection of the node point onto the corresponding axis."""
         return self.projector(self.point)
 
-    def distance_to_point(self, point: Point) -> Coordinate:
+    def distance_to_point(self, point: _Point) -> _Coordinate:
         """Calculates distance to given point."""
-        return _points_distance(self.point, point)
+        return self.metric(self.point, point)
 
-    def distance_to_coordinate(self, coordinate: Coordinate) -> Coordinate:
+    def distance_to_coordinate(self, coordinate: _Coordinate) -> _Coordinate:
         """Calculates distance to given coordinate."""
         return (self.projection - coordinate) ** 2
 
@@ -74,11 +71,13 @@ class Tree:
         https://en.wikipedia.org/wiki/K-d_tree
     """
 
-    __slots__ = '_points', '_root'
+    __slots__ = '_context', '_points', '_root'
 
-    def __init__(self, points: Sequence[Point],
+    def __init__(self,
+                 points: _Sequence[_Point],
                  *,
-                 node_cls: Optional[Type[Node]] = None) -> None:
+                 node_cls: _Type[Node] = Node,
+                 context: _Optional[_Context] = None) -> None:
         """
         Initializes tree from points.
 
@@ -95,26 +94,29 @@ class Tree:
         >>> points = list(map(Point, range(-5, 6), range(10)))
         >>> tree = Tree(points)
         """
-        self._points = points
-        self._root = _create_node(Node if node_cls is None else node_cls,
-                                  range(len(points)), points, False)
+        if context is None:
+            context = _get_context()
+        self._context, self._points, self._root = (
+            context, points, _create_node(node_cls, range(len(points)), points,
+                                          False,
+                                          context.points_squared_distance))
 
-    __repr__ = generate_repr(__init__)
+    __repr__ = _generate_repr(__init__)
 
     @property
-    def node_cls(self) -> Type[Node]:
+    def context(self) -> _Context:
         """
-        Returns type of the nodes.
+        Returns context of the tree.
 
         Time complexity:
             ``O(1)``
         Memory complexity:
             ``O(1)``
         """
-        return type(self._root)
+        return self._context
 
     @property
-    def points(self) -> Sequence[Point]:
+    def points(self) -> _Sequence[_Point]:
         """
         Returns underlying points.
 
@@ -133,7 +135,7 @@ class Tree:
         """
         return self._points
 
-    def n_nearest_indices(self, n: int, point: Point) -> Sequence[int]:
+    def n_nearest_indices(self, n: int, point: _Point) -> _Sequence[int]:
         """
         Searches for indices of points in the tree
         that are the nearest to the given point.
@@ -167,7 +169,7 @@ class Tree:
                 if n < len(self._points)
                 else range(len(self._points)))
 
-    def n_nearest_points(self, n: int, point: Point) -> Sequence[Point]:
+    def n_nearest_points(self, n: int, point: _Point) -> _Sequence[_Point]:
         """
         Searches for points in the tree the nearest to the given point.
 
@@ -200,7 +202,7 @@ class Tree:
                 if n < len(self._points)
                 else self._points)
 
-    def n_nearest_items(self, n: int, point: Point) -> Sequence[Item]:
+    def n_nearest_items(self, n: int, point: _Point) -> _Sequence[Item]:
         """
         Searches for indices with points in the tree
         that are the nearest to the given point.
@@ -236,7 +238,7 @@ class Tree:
                 if n < len(self._points)
                 else list(enumerate(self._points)))
 
-    def _n_nearest_items(self, n: int, point: Point) -> List[Item]:
+    def _n_nearest_items(self, n: int, point: _Point) -> _List[Item]:
         candidates = []  # type: List[Tuple[Coordinate, Item]]
         queue = [self._root]
         push, pop = queue.append, queue.pop
@@ -245,9 +247,9 @@ class Tree:
             distance_to_point = node.distance_to_point(point)
             candidate = -distance_to_point, node.item
             if len(candidates) < n:
-                heappush(candidates, candidate)
+                _heappush(candidates, candidate)
             elif distance_to_point < -candidates[0][0]:
-                heapreplace(candidates, candidate)
+                _heapreplace(candidates, candidate)
             coordinate = node.projector(point)
             point_is_on_the_left = coordinate < node.projection
             if point_is_on_the_left:
@@ -265,7 +267,7 @@ class Tree:
                     push(node.left)
         return [item for _, item in candidates]
 
-    def nearest_index(self, point: Point) -> int:
+    def nearest_index(self, point: _Point) -> int:
         """
         Searches for index of a point in the tree
         that is the nearest to the given point.
@@ -296,7 +298,7 @@ class Tree:
         result, _ = self.nearest_item(point)
         return result
 
-    def nearest_point(self, point: Point) -> Point:
+    def nearest_point(self, point: _Point) -> _Point:
         """
         Searches for point in the tree that is the nearest to the given point.
 
@@ -326,7 +328,7 @@ class Tree:
         _, result = self.nearest_item(point)
         return result
 
-    def nearest_item(self, point: Point) -> Item:
+    def nearest_item(self, point: _Point) -> Item:
         """
         Searches for index with point in the tree
         that is the nearest to the given point.
@@ -378,7 +380,7 @@ class Tree:
                     push(node.left)
         return result
 
-    def find_box_indices(self, box: Box) -> List[int]:
+    def find_box_indices(self, box: _Box) -> _List[int]:
         """
         Searches for indices of points that lie inside the given box.
 
@@ -410,7 +412,7 @@ class Tree:
         """
         return [index for index, _ in self._find_box_items(box)]
 
-    def find_box_points(self, box: Box) -> List[Point]:
+    def find_box_points(self, box: _Box) -> _List[_Point]:
         """
         Searches for points that lie inside the given box.
 
@@ -443,7 +445,7 @@ class Tree:
         """
         return [point for _, point in self._find_box_items(box)]
 
-    def find_box_items(self, box: Box) -> List[Item]:
+    def find_box_items(self, box: _Box) -> _List[Item]:
         """
         Searches for indices with points in the tree
         that lie inside the given box.
@@ -477,7 +479,7 @@ class Tree:
         """
         return list(self._find_box_items(box))
 
-    def _find_box_items(self, box: Box) -> Iterator[Item]:
+    def _find_box_items(self, box: _Box) -> _Iterator[Item]:
         queue = [self._root]
         push, pop = queue.append, queue.pop
         while queue:
@@ -494,10 +496,12 @@ class Tree:
                 push(node.right)
 
 
-def _create_node(cls: Type[Node],
-                 indices: Sequence[int],
-                 points: Sequence[Point],
-                 is_y_axis: bool) -> Union[Node, NIL]:
+def _create_node(cls: _Type[Node],
+                 indices: _Sequence[int],
+                 points: _Sequence[_Point],
+                 is_y_axis: bool,
+                 metric: _Callable[[_Point, _Point], _Coordinate]
+                 ) -> _Union[Node, NIL]:
     if not indices:
         return NIL
     indices = sorted(indices,
@@ -508,6 +512,6 @@ def _create_node(cls: Type[Node],
     next_is_y_axis = not is_y_axis
     return cls(pivot_index, points[pivot_index], is_y_axis,
                _create_node(cls, indices[:middle_index], points,
-                            next_is_y_axis),
+                            next_is_y_axis, metric),
                _create_node(cls, indices[middle_index + 1:], points,
-                            next_is_y_axis))
+                            next_is_y_axis, metric), metric)
