@@ -1,74 +1,84 @@
-from functools import reduce
-from heapq import (heappop,
-                   heappush)
-from math import (floor,
-                  inf)
-from typing import (Iterator,
-                    Optional,
-                    Sequence,
-                    Tuple,
-                    Type)
+from functools import reduce as _reduce
+from heapq import (heappop as _heappop,
+                   heappush as _heappush)
+from math import (floor as _floor,
+                  inf as _inf)
+from typing import (Callable as _Callable,
+                    Iterator as _Iterator,
+                    Optional as _Optional,
+                    Sequence as _Sequence,
+                    Tuple as _Tuple)
 
 from ground.base import (Context as _Context,
                          get_context as _get_context)
-from ground.hints import (Box,
-                          Coordinate,
-                          Point,
-                          Segment)
-from reprit.base import generate_repr
+from ground.hints import (Box as _Box,
+                          Coordinate as _Coordinate,
+                          Point as _Point,
+                          Segment as _Segment)
+from reprit.base import generate_repr as _generate_repr
 
-from .core import (box as _box,
-                   hilbert as _hilbert,
-                   segment as _segment)
-from .core.utils import ceil_division
+from .core import hilbert as _hilbert
+from .core.utils import ceil_division as _ceil_division
 
-try:
-    from typing import Protocol
-except ImportError:
-    from typing_extensions import Protocol
-
-Item = Tuple[int, Segment]
+Item = _Tuple[int, _Segment]
 
 
-class Node(Protocol):
-    """
-    Interface of segmental *R*-tree node.
+class Node:
+    """Represents node of segmental *R*-tree."""
+    __slots__ = ('box', 'box_point_metric', 'box_segment_metric',
+                 'children', 'index', 'segment', 'segment_point_metric',
+                 'segments_metric')
 
-    Can be implemented for custom metrics definition.
-    """
+    def __init__(self,
+                 index: int,
+                 box: _Box,
+                 segment: _Optional[_Segment],
+                 children: _Optional[_Sequence['Node']],
+                 box_point_metric: _Callable[[_Box, _Point], _Coordinate],
+                 box_segment_metric
+                 : _Callable[[_Box, _Point, _Point], _Coordinate],
+                 segment_point_metric
+                 : _Callable[[_Point, _Point, _Point], _Coordinate],
+                 segments_metric
+                 : _Callable[[_Point, _Point, _Point, _Point], _Coordinate]
+                 ) -> None:
+        self.box, self.children, self.index, self.segment = (
+            box, children, index, segment)
+        self.box_point_metric, self.box_segment_metric = (
+            box_point_metric, box_segment_metric)
+        self.segment_point_metric, self.segments_metric = (
+            segment_point_metric, segments_metric)
 
-    def __new__(cls,
-                index: int,
-                box: Box,
-                segment: Optional[Segment],
-                children: Optional[Sequence['Node']]) -> 'Node':
-        """Creates node."""
-
-    @property
-    def box(self) -> Box:
-        """Returns box of the node."""
-
-    @property
-    def children(self) -> Sequence['Node']:
-        """Returns children of the node."""
-
-    @property
-    def index(self) -> int:
-        """Returns index of the node."""
+    __repr__ = _generate_repr(__init__)
 
     @property
     def is_leaf(self) -> bool:
-        """Checks whether the node is a leaf."""
+        return self.children is None
 
     @property
     def item(self) -> Item:
-        """Returns underlying index with segment."""
+        return self.index, self.segment
 
-    def distance_to_point(self, point: Point) -> Coordinate:
-        """Calculates distance to given point."""
+    def distance_to_point(self,
+                          point: _Point,
+                          *,
+                          _minus_inf: _Coordinate = -_inf) -> _Coordinate:
+        return (self.segment_point_metric(self.segment.start,
+                                          self.segment.end, point)
+                or _minus_inf
+                if self.is_leaf
+                else self.box_point_metric(self.box, point))
 
-    def distance_to_segment(self, segment: Segment) -> Coordinate:
-        """Calculates distance to given segment."""
+    def distance_to_segment(self,
+                            segment: _Segment,
+                            *,
+                            _minus_inf: _Coordinate = -_inf) -> _Coordinate:
+        return (self.segments_metric(self.segment.start, self.segment.end,
+                                     segment.start, segment.end)
+                or _minus_inf
+                if self.is_leaf
+                else self.box_segment_metric(self.box, segment.start,
+                                             segment.end))
 
 
 class Tree:
@@ -78,13 +88,13 @@ class Tree:
     Reference:
         https://en.wikipedia.org/wiki/Hilbert_R-tree#Packed_Hilbert_R-trees
     """
-    __slots__ = '_segments', '_max_children', '_root'
+    __slots__ = '_context', '_max_children', '_root', '_segments'
 
     def __init__(self,
-                 segments: Sequence[Segment],
+                 segments: _Sequence[_Segment],
                  *,
                  max_children: int = 16,
-                 node_cls: Optional[Type[Node]] = None) -> None:
+                 context: _Optional[_Context] = None) -> None:
         """
         Initializes tree from segments.
 
@@ -103,49 +113,39 @@ class Tree:
         ...             for index in range(1, 11)]
         >>> tree = Tree(segments)
         """
-        self._segments = segments
-        self._max_children = max_children
-        context = _get_context()
-        self._root = _create_root(segments, max_children,
-                                  _to_default_node_cls(context)
-                                  if node_cls is None
-                                  else node_cls,
-                                  context)
+        if context is None:
+            context = _get_context()
+        box_cls = context.box_cls
+        self._context, self._max_children, self._root, self._segments = (
+            context, max_children,
+            _create_root(segments,
+                         [box_cls(*((segment.start.x, segment.end.x)
+                                    if segment.start.x < segment.end.x
+                                    else (segment.end.x, segment.start.x)),
+                                  *((segment.start.y, segment.end.y)
+                                    if segment.start.y < segment.end.y
+                                    else (segment.end.y, segment.start.y)))
+                          for segment in segments], max_children,
+                         context.merged_box,
+                         context.box_point_squared_distance,
+                         context.box_segment_squared_distance,
+                         context.segment_point_squared_distance,
+                         context.segments_squared_distance),
+            segments)
 
-    __repr__ = generate_repr(__init__)
+    __repr__ = _generate_repr(__init__)
 
     @property
-    def segments(self) -> Sequence[Segment]:
+    def context(self) -> _Context:
         """
-        Returns underlying segments.
-
-        Time complexity:
-            ``O(1)``
-        Memory complexity:
-            ``O(1)``
-
-        >>> from ground.base import get_context
-        >>> context = get_context()
-        >>> Point, Segment = context.point_cls, context.segment_cls
-        >>> segments = [Segment(Point(0, index), Point(index, index))
-        ...             for index in range(1, 11)]
-        >>> tree = Tree(segments)
-        >>> tree.segments == segments
-        True
-        """
-        return self._segments
-
-    @property
-    def node_cls(self) -> Type[Node]:
-        """
-        Returns type of the nodes.
+        Returns context of the tree.
 
         Time complexity:
             ``O(1)``
         Memory complexity:
             ``O(1)``
         """
-        return type(self._root)
+        return self._context
 
     @property
     def max_children(self) -> int:
@@ -168,7 +168,28 @@ class Tree:
         """
         return self._max_children
 
-    def n_nearest_indices(self, n: int, segment: Segment) -> Sequence[int]:
+    @property
+    def segments(self) -> _Sequence[_Segment]:
+        """
+        Returns underlying segments.
+
+        Time complexity:
+            ``O(1)``
+        Memory complexity:
+            ``O(1)``
+
+        >>> from ground.base import get_context
+        >>> context = get_context()
+        >>> Point, Segment = context.point_cls, context.segment_cls
+        >>> segments = [Segment(Point(0, index), Point(index, index))
+        ...             for index in range(1, 11)]
+        >>> tree = Tree(segments)
+        >>> tree.segments == segments
+        True
+        """
+        return self._segments
+
+    def n_nearest_indices(self, n: int, segment: _Segment) -> _Sequence[int]:
         """
         Searches for indices of segments in the tree
         the nearest to the given segment.
@@ -205,7 +226,7 @@ class Tree:
                 if n < len(self._segments)
                 else range(len(self._segments)))
 
-    def n_nearest_items(self, n: int, segment: Segment) -> Sequence[Item]:
+    def n_nearest_items(self, n: int, segment: _Segment) -> _Sequence[Item]:
         """
         Searches for indices with segments in the tree
         the nearest to the given segment.
@@ -244,8 +265,8 @@ class Tree:
                     if n < len(self._segments)
                     else enumerate(self._segments))
 
-    def n_nearest_segments(self, n: int, segment: Segment
-                           ) -> Sequence[Segment]:
+    def n_nearest_segments(self, n: int, segment: _Segment
+                           ) -> _Sequence[_Segment]:
         """
         Searches for segments in the tree the nearest to the given segment.
 
@@ -281,8 +302,8 @@ class Tree:
                 if n < len(self._segments)
                 else self._segments)
 
-    def n_nearest_to_point_indices(self, n: int, point: Point
-                                   ) -> Sequence[int]:
+    def n_nearest_to_point_indices(self, n: int, point: _Point
+                                   ) -> _Sequence[int]:
         """
         Searches for indices of segments in the tree
         the nearest to the given point.
@@ -319,7 +340,8 @@ class Tree:
                 if n < len(self._segments)
                 else range(len(self._segments)))
 
-    def n_nearest_to_point_items(self, n: int, point: Point) -> Sequence[Item]:
+    def n_nearest_to_point_items(self, n: int, point: _Point) -> _Sequence[
+        Item]:
         """
         Searches for indices with segments in the tree
         the nearest to the given point.
@@ -358,8 +380,8 @@ class Tree:
                     if n < len(self._segments)
                     else enumerate(self._segments))
 
-    def n_nearest_to_point_segments(self, n: int, point: Point
-                                    ) -> Sequence[Segment]:
+    def n_nearest_to_point_segments(self, n: int, point: _Point
+                                    ) -> _Sequence[_Segment]:
         """
         Searches for segments in the tree the nearest to the given point.
 
@@ -395,7 +417,7 @@ class Tree:
                 if n < len(self._segments)
                 else self._segments)
 
-    def nearest_index(self, segment: Segment) -> int:
+    def nearest_index(self, segment: _Segment) -> int:
         """
         Searches for index of segment in the tree
         the nearest to the given segment.
@@ -424,7 +446,7 @@ class Tree:
         result, _ = self.nearest_item(segment)
         return result
 
-    def nearest_item(self, segment: Segment) -> Item:
+    def nearest_item(self, segment: _Segment) -> Item:
         """
         Searches for index with segment in the tree
         the nearest to the given segment.
@@ -453,17 +475,17 @@ class Tree:
         """
         queue = [(0, 0, self._root)]
         while queue:
-            _, _, node = heappop(queue)
+            _, _, node = _heappop(queue)
             for child in node.children:
-                heappush(queue,
-                         (child.distance_to_segment(segment),
-                          child.index if child.is_leaf else -child.index - 1,
-                          child))
+                _heappush(queue,
+                          (child.distance_to_segment(segment),
+                           child.index if child.is_leaf else -child.index - 1,
+                           child))
             if queue and queue[0][1] >= 0:
-                _, _, node = heappop(queue)
+                _, _, node = _heappop(queue)
                 return node.item
 
-    def nearest_segment(self, segment: Segment) -> Segment:
+    def nearest_segment(self, segment: _Segment) -> _Segment:
         """
         Searches for segment in the tree the nearest to the given segment.
 
@@ -491,7 +513,7 @@ class Tree:
         _, result = self.nearest_item(segment)
         return result
 
-    def nearest_to_point_index(self, point: Point) -> int:
+    def nearest_to_point_index(self, point: _Point) -> int:
         """
         Searches for index of segment in the tree
         the nearest to the given point.
@@ -519,7 +541,7 @@ class Tree:
         result, _ = self.nearest_to_point_item(point)
         return result
 
-    def nearest_to_point_item(self, point: Point) -> Item:
+    def nearest_to_point_item(self, point: _Point) -> Item:
         """
         Searches for index with segment in the tree
         the nearest to the given point.
@@ -548,17 +570,17 @@ class Tree:
         """
         queue = [(0, 0, self._root)]
         while queue:
-            _, _, node = heappop(queue)
+            _, _, node = _heappop(queue)
             for child in node.children:
-                heappush(queue,
-                         (child.distance_to_point(point),
-                          child.index if child.is_leaf else -child.index - 1,
-                          child))
+                _heappush(queue,
+                          (child.distance_to_point(point),
+                           child.index if child.is_leaf else -child.index - 1,
+                           child))
             if queue and queue[0][1] >= 0:
-                _, _, node = heappop(queue)
+                _, _, node = _heappop(queue)
                 return node.item
 
-    def nearest_to_point_segment(self, point: Point) -> Segment:
+    def nearest_to_point_segment(self, point: _Point) -> _Segment:
         """
         Searches for segment in the tree the nearest to the given point.
 
@@ -586,80 +608,82 @@ class Tree:
         _, result = self.nearest_to_point_item(point)
         return result
 
-    def _n_nearest_items(self, n: int, segment: Segment) -> Iterator[Item]:
+    def _n_nearest_items(self, n: int, segment: _Segment) -> _Iterator[Item]:
         queue = [(0, 0, self._root)]
         while n and queue:
-            _, _, node = heappop(queue)
+            _, _, node = _heappop(queue)
             for child in node.children:
-                heappush(queue,
-                         (child.distance_to_segment(segment),
-                          child.index if child.is_leaf else -child.index - 1,
-                          child))
+                _heappush(queue,
+                          (child.distance_to_segment(segment),
+                           child.index if child.is_leaf else -child.index - 1,
+                           child))
             while n and queue and queue[0][1] >= 0:
-                _, _, node = heappop(queue)
+                _, _, node = _heappop(queue)
                 yield node.item
                 n -= 1
 
-    def _n_nearest_to_point_items(self, n: int, point: Point
-                                  ) -> Iterator[Item]:
+    def _n_nearest_to_point_items(self, n: int, point: _Point
+                                  ) -> _Iterator[Item]:
         queue = [(0, 0, self._root)]
         while n and queue:
-            _, _, node = heappop(queue)
+            _, _, node = _heappop(queue)
             for child in node.children:
-                heappush(queue,
-                         (child.distance_to_point(point),
-                          child.index if child.is_leaf else -child.index - 1,
-                          child))
+                _heappush(queue,
+                          (child.distance_to_point(point),
+                           child.index if child.is_leaf else -child.index - 1,
+                           child))
             while n and queue and queue[0][1] >= 0:
-                _, _, node = heappop(queue)
+                _, _, node = _heappop(queue)
                 yield node.item
                 n -= 1
 
 
-def _create_root(segments: Sequence[Segment],
+def _create_root(segments: _Sequence[_Segment],
+                 boxes: _Sequence[_Box],
                  max_children: int,
-                 node_cls: Type[Node],
-                 context: _Context) -> Node:
-    box_cls, merge_boxes = context.box_cls, context.merged_box
-    boxes = [box_cls(*((segment.start.x, segment.end.x)
-                       if segment.start.x < segment.end.x
-                       else (segment.end.x, segment.start.x)),
-                     *((segment.start.y, segment.end.y)
-                       if segment.start.y < segment.end.y
-                       else (segment.end.y, segment.start.y)))
-             for segment in segments]
-    nodes = [node_cls(index, box, segment, None)
+                 boxes_merger: _Callable[[_Box, _Box], _Box],
+                 box_point_metric: _Callable[[_Box, _Point], _Coordinate],
+                 box_segment_metric: _Callable[
+                     [_Box, _Point, _Point], _Coordinate],
+                 segment_point_metric
+                 : _Callable[[_Point, _Point, _Point], _Coordinate],
+                 segments_metric
+                 : _Callable[
+                     [_Point, _Point, _Point, _Point], _Coordinate]) -> Node:
+    nodes = [Node(index, box, segment, None, box_point_metric,
+                  box_segment_metric, segment_point_metric, segments_metric)
              for index, (box, segment) in enumerate(zip(boxes, segments))]
-    root_box = reduce(merge_boxes, boxes)
+    root_box = _reduce(boxes_merger, boxes)
     leaves_count = len(nodes)
     if leaves_count <= max_children:
         # only one node, skip sorting and just fill the root box
-        return node_cls(len(nodes), root_box, None, nodes)
+        return Node(len(nodes), root_box, None, nodes, box_point_metric,
+                    box_segment_metric, segment_point_metric, segments_metric)
     else:
         def node_key(node: Node,
-                     double_root_delta_x: Coordinate
+                     double_root_delta_x: _Coordinate
                      = 2 * (root_box.max_x - root_box.min_x) or 1,
-                     double_root_delta_y: Coordinate
+                     double_root_delta_y: _Coordinate
                      = 2 * (root_box.max_y - root_box.min_y) or 1,
-                     double_root_min_x: Coordinate = 2 * root_box.min_x,
-                     double_root_min_y: Coordinate = 2 * root_box.min_y
+                     double_root_min_x: _Coordinate = 2 * root_box.min_x,
+                     double_root_min_y: _Coordinate = 2 * root_box.min_y
                      ) -> int:
             box = node.box
-            return _hilbert.index(floor(_hilbert.MAX_COORDINATE
-                                        * (box.min_x + box.max_x
-                                           - double_root_min_x)
-                                        / double_root_delta_x),
-                                  floor(_hilbert.MAX_COORDINATE
-                                        * (box.min_y + box.max_y
-                                           - double_root_min_y)
-                                        / double_root_delta_y))
+            return _hilbert.index(_floor(_hilbert.MAX_COORDINATE
+                                         * (box.min_x + box.max_x
+                                            - double_root_min_x)
+                                         / double_root_delta_x),
+                                  _floor(_hilbert.MAX_COORDINATE
+                                         * (box.min_y + box.max_y
+                                            - double_root_min_y)
+                                         / double_root_delta_y))
 
         nodes = sorted(nodes,
                        key=node_key)
         nodes_count = step = leaves_count
         levels_limits = [nodes_count]
         while True:
-            step = ceil_division(step, max_children)
+            step = _ceil_division(step, max_children)
             if step == 1:
                 break
             nodes_count += step
@@ -669,63 +693,11 @@ def _create_root(segments: Sequence[Segment],
             while start < level_limit:
                 stop = min(start + max_children, level_limit)
                 children = nodes[start:stop]
-                nodes.append(node_cls(len(nodes),
-                                      reduce(merge_boxes,
-                                             [child.box
-                                              for child in children]),
-                                      None, children))
+                nodes.append(Node(len(nodes),
+                                  _reduce(boxes_merger,
+                                          [child.box for child in children]),
+                                  None, children, box_point_metric,
+                                  box_segment_metric, segment_point_metric,
+                                  segments_metric))
                 start = stop
         return nodes[-1]
-
-
-def _to_default_node_cls(context: _Context) -> Type[Node]:
-    class Node:
-        __slots__ = 'index', 'box', 'segment', 'children'
-
-        def __init__(self,
-                     index: int,
-                     box: Box,
-                     segment: Optional[Segment],
-                     children: Optional[Sequence['Node']]) -> None:
-            self.index, self.box, self.segment, self.children = (
-                index, box, segment, children)
-
-        __repr__ = generate_repr(__init__)
-
-        _dot_product = staticmethod(context.dot_product)
-        _point_cls = staticmethod(context.point_cls)
-        _segments_relation = staticmethod(context.segments_relation)
-
-        @property
-        def is_leaf(self) -> bool:
-            return self.children is None
-
-        @property
-        def item(self) -> Item:
-            return self.index, self.segment
-
-        def distance_to_point(self, point: Point,
-                              *,
-                              _minus_inf: Coordinate = -inf) -> Coordinate:
-            return (_segment.distance_to_point(self._dot_product,
-                                               self.segment.start,
-                                               self.segment.end, point)
-                    or _minus_inf
-                    if self.is_leaf
-                    else _box.distance_to_point(self.box, point))
-
-        def distance_to_segment(self, segment: Segment,
-                                *,
-                                _minus_inf: Coordinate = -inf) -> Coordinate:
-            return (_segment.distance_to(self._dot_product,
-                                         self._segments_relation,
-                                         self.segment.start, self.segment.end,
-                                         segment.start, segment.end)
-                    or _minus_inf
-                    if self.is_leaf
-                    else _segment.distance_to_box(self._dot_product,
-                                                  self._point_cls,
-                                                  self._segments_relation,
-                                                  segment, self.box))
-
-    return Node
